package Physics;

//import com.jogamp.opengl.GL;
//import com.jogamp.opengl.GLAutoDrawable;

import javax.media.opengl.GL;
import javax.media.opengl.GLAutoDrawable;

/**
 * http://ironphoenix.org/tril/bif/
 */
class SpiralGLView extends BaseGLView 
{
	private static final long serialVersionUID = 1L;
	static SpiralSimulation _sim;
	
	public SpiralGLView( SpiralSimulation aSimulation )
	{
//		super( aSimulation );
		_sim = aSimulation;
		_PositionZ = -4.0;
	}

	/**
	 * The distortion is divided to the other two room-dimensions!
	 * Only a rotating structure can be stable.
	 * @param rdLastRadius: radius from the last calculation-step
	 * @param rdActualRadius: radius from actual calculation-step
	 */
	double CalcSpiralLength( double rdLastRadius, double rdActualRadius )
	{
		double dDistortLast = 1.0 / (rdLastRadius ); //Radius^2 center to neighbor position
		double dDistortActual = 1.0 / (rdActualRadius ); //Radius^2 center to actual position

		// length = radius
		// length = sqrt(   ( r1 - r2 )^2   +   ( 1/r1 - 1/r2 )^2   )
		return (rdActualRadius-rdLastRadius) * (rdActualRadius-rdLastRadius)
			   + (dDistortActual-dDistortLast) * (dDistortActual-dDistortLast);	
	}

	/**
	 * The distortion is divided to the other two room-dimensions!
	 * Only a rotating structure can be stable.
	 * @param dShiftZ: the minimum distance from a plain spiral to the center of mass
	 * @param rdLastRadius: radius from the last calculation-step
	 * @param rdActualRadius: radius from actual calculation-step
	 */
	double CalcSpiralLengthSquare( double dShiftZ, double rdLastRadius, double rdActualRadius  )
	{
		double dRadiusSquareLast = rdLastRadius*rdLastRadius;
		double dRadiusSquareActual = rdActualRadius*rdActualRadius;
		double dDistortLast = 1.0 / (dRadiusSquareLast + dShiftZ*dShiftZ ); //Radius^2 center to neighbor position
		double dDistortActual = 1.0 / (dRadiusSquareActual + dShiftZ*dShiftZ ); //Radius^2 center to actual position

		// length = radius
		// length = sqrt(   ( r1 - r2 )^2   +   ( 1/r1^2 - 1/r2^2 )^2   )
		return (rdActualRadius-rdLastRadius) * (rdActualRadius-rdLastRadius)
			   + (dDistortActual-dDistortLast) * (dDistortActual-dDistortLast);	
	}
	
	/**
	 * Calculate length alpha and beta with cosines-theorem.
	 * Parameter:
	 *   rdLastRadius: radius from the last calculation-step
	 *   rdActualRadius: radius from actual calculation-step
	 *   rdLengthSquare: the length of the spiral 
	 */
	double CalcSpiralAlpha( double rdLastRadius, double rdActualRadius, double rdLengthSquare )
	{
		double dRadiusSquareLast = rdLastRadius*rdLastRadius;
		double dRadiusSquareActual = rdActualRadius*rdActualRadius;

		//cos(alpha) = 1/2 * ( rl^2 + ra^2 - dL^2 ) / ( rl * ra ); //law of cosines
		return (dRadiusSquareLast + dRadiusSquareActual - rdLengthSquare) 
				/ ( 2.0 * Math.sqrt( dRadiusSquareLast * dRadiusSquareActual ) );
	}
	double CalcSpiralBeta( double rdLastRadius, double rdActualRadius, double rdLengthSquare )
	{
		double dRadiusSquareLast = rdLastRadius*rdLastRadius;
		double dRadiusSquareActual = rdActualRadius*rdActualRadius;
	
		//cos(beta) = 1/2 * ( dL^2 + rl^2 - ra^2 ) / ( dL * rl ); //with law of cosines
		return (rdLengthSquare + dRadiusSquareLast - dRadiusSquareActual ) 
				 / ( 2.0 * Math.sqrt( rdLengthSquare ) * rdLastRadius );
	}	

	/**
	 * spiral generation
	 */
	void OnSpiral2D( boolean bModeSquare )
	{
		//initialize spiral
		int nPoints = 5000; //number of points
		double dStep = 1.0/nPoints; //number of points
		double dActualRadius = 5.0; //initial distance
		double dMinRadius = _sim.dSliderPos/100; //abort calculation at minimal radius 
		double dLastRadius = dActualRadius-dStep; //radius of previous step

		Vector3d oDirection = new Vector3d( -1.0, 0.0, 0.0 ); //direction of grid x-direction	
		Vector3d oPosition = new Vector3d( dActualRadius, 0.0, 0.0 );
		double dBetaLast = 0.0;

		gl.glBegin( GL.GL_LINE_STRIP );
		gl.glColor4d(1.0, 1.0, 1.0, 1.0); //white

		for(; dActualRadius>dMinRadius; dActualRadius-=dStep ) 
		{
			double dCosAlpha = 0.0;
			double dCosBeta = 0.0;
			double dLengthSquare = 0.0;

			if( bModeSquare )
			{
				dLengthSquare = CalcSpiralLengthSquare( 0.0, dLastRadius, dActualRadius );
				dCosAlpha = CalcSpiralAlpha( dLastRadius, dActualRadius, dLengthSquare );
				dCosBeta = CalcSpiralBeta( dLastRadius, dActualRadius, dLengthSquare );			
			}
			else
			{
				dLengthSquare = CalcSpiralLength( dLastRadius, dActualRadius );
				dCosAlpha = CalcSpiralAlpha( dLastRadius, dActualRadius, dLengthSquare );
				dCosBeta = CalcSpiralBeta( dLastRadius, dActualRadius, dLengthSquare );			
			}

			double dBeta = Math.acos( dCosBeta ) * 180.0/Math.PI; //diagonal length is sqrt(2)
			double dAlpha = Math.acos( dCosAlpha ) * 180.0/Math.PI; //diagonal length is sqrt(2)

			//oRotation shows into from oDirection-vector always to the center.
			//and applying to this we rotate sqrt(2) in oSphereTangent direction
			//and sqrt(2) perpendicular to oSphereTangent into direction oCrossSphereTangent
			Vector3d oSphereTangent = oPosition.crossProduct( oDirection ); //always perpendicular to oDirection
			if( oSphereTangent.absoluteValue() < 0.0001 ) //TODO MAX_NULL_TOLERANZ
				oSphereTangent.setComponents( 0.0, 0.0, 1.0 ); 
			oSphereTangent.divEqu( oSphereTangent.absoluteValue() ); //get vector with length 1

			Vector3d oCrossSphereTangent = oSphereTangent.crossProduct( oDirection ); //perpendicular to oRotation and oDirection
			oCrossSphereTangent.divEqu( oCrossSphereTangent.absoluteValue() ); //get vector with length 1

			// rotate because of distortion
			oDirection.rotateX( 1.0 * oSphereTangent.getX() * (dBetaLast-dBeta) );
			oDirection.rotateY( 1.0 * oSphereTangent.getY() * (dBetaLast-dBeta) );
			oDirection.rotateZ( 1.0 * oSphereTangent.getZ() * (dBetaLast-dBeta) );
			dBetaLast = dBeta;

			// rotate around center
			oDirection.rotateX( 1.0 * oSphereTangent.getX() * dAlpha );
			oDirection.rotateY( 1.0 * oSphereTangent.getY() * dAlpha );
			oDirection.rotateZ( 1.0 * oSphereTangent.getZ() * dAlpha );

			//calculate position of spiral
			oPosition.addEqu(  oDirection.mul( Math.sqrt(dLengthSquare) )  );
			dLastRadius = dActualRadius; //save the last radius-state.

			//correction radius 
//			oPosition *= dActualRadius*dActualRadius / oPosition.AbsoluteValueSquare();

			gl.glVertex3d(oPosition.getX()*1.0, oPosition.getY()*1.0, oPosition.getZ()*1.0 );	
		}

		gl.glEnd();
	}

	/**
	 * Spiral rotation around axis!
	 * @param nIndex: Mirror spiral with this nIndex (0..5 sides of a cube) 
	 *  beta += acos( dCosAlpha )*500; //Scheibengalaxie?
	 *  beta += 0.02 and dGamma+=0.01 //Elektronenorbitale?
	 */
	void OnSpiral3D( boolean bModeSquare, boolean bMode2D, int nIndex, boolean constRadius )
	{	
		double dShiftZ = 0.0;
		//Set a minimum distance to the center of mass.
		double dMinCenterDist = _sim.dSliderPos/100.0; //0.12; //abort calculation at minimal radius
		double dMinRadius = 0.0;
		if( dMinCenterDist*dMinCenterDist > dShiftZ*dShiftZ ) //we must not go until zero-radius
			dMinRadius = Math.sqrt( dMinCenterDist*dMinCenterDist - dShiftZ*dShiftZ );

		//initialize spiral
		double dAlpha = 0.0, dBeta=0.0, dGamma=0.0; //spiral rotation around z-axis
		//double dLastAlpha = 0.0;

		double dStep = 1.0/4000.0; //1000; //number of points
		double dLastRadius = 5.0; //initial distance

		//int nStep = 0;
		//boolean bTearOffReached = false;
		Vector3d oLastPosition = new Vector3d( constRadius?1.0:dLastRadius, 0.0, 0.0 ); //the last direction vector

		gl.glBegin( GL.GL_LINE_STRIP );
		gl.glColor4d(1.0, 1.0, 1.0, 1.0); //white

		for( double dActualRadius=dLastRadius-dStep; dActualRadius>dMinRadius; dActualRadius-=dStep ) 
		{
			double dCosAlpha=0.0;
			//double dCosBeta = 0.0;  //not used!
			double dLengthSquare = 0.0; //not used!
			Vector3d oPosition = new Vector3d( constRadius?1.0:dLastRadius, 0.0, dShiftZ );

			if( bMode2D )
			{
				if( bModeSquare )
				{
					dLengthSquare = CalcSpiralLengthSquare( dShiftZ, dLastRadius, dActualRadius );
					dCosAlpha = CalcSpiralAlpha( dLastRadius, dActualRadius, dLengthSquare );
					//dCosBeta = CalcSpiralBeta( dLastRadius, dActualRadius, dLengthSquare );			
				}
				else
				{
					dLengthSquare = CalcSpiralLength( dLastRadius, dActualRadius );
					dCosAlpha = CalcSpiralAlpha( dLastRadius, dActualRadius, dLengthSquare );
					//dCosBeta = CalcSpiralBeta( dLastRadius, dActualRadius, dLengthSquare );			
				}

				dAlpha += Math.acos( dCosAlpha );
				oPosition.rotateZ( dAlpha*180.0/Math.PI );
			}
			else
			{
				if( bModeSquare )
				{
					dLengthSquare = CalcSpiralLengthSquare( oPosition.getX(), dLastRadius, dActualRadius );
					dCosAlpha = CalcSpiralAlpha( dLastRadius, dActualRadius, dLengthSquare );
					//dCosBeta = CalcSpiralBeta( dLastRadius, dActualRadius, dLengthSquare );			
				}
				else
				{
					dLengthSquare = CalcSpiralLength( dLastRadius, dActualRadius );
					dCosAlpha = CalcSpiralAlpha( dLastRadius, dActualRadius, dLengthSquare );
					//dCosBeta = CalcSpiralBeta( dLastRadius, dActualRadius, dLengthSquare );			
				}

				dAlpha += Math.acos( dCosAlpha ); //* 1/sqrt(2)????
				oPosition.rotateX( dAlpha*180.0/Math.PI );

				if( bModeSquare )
				{
					dLengthSquare = CalcSpiralLengthSquare( oPosition.getY(), dLastRadius, dActualRadius );
					dCosAlpha = CalcSpiralAlpha( dLastRadius, dActualRadius, dLengthSquare );
					//dCosBeta = CalcSpiralBeta( dLastRadius, dActualRadius, dLengthSquare );			
				}
				else
				{
					dLengthSquare = CalcSpiralLength( dLastRadius, dActualRadius );
					dCosAlpha = CalcSpiralAlpha( dLastRadius, dActualRadius, dLengthSquare );
					//dCosBeta = CalcSpiralBeta( dLastRadius, dActualRadius, dLengthSquare );			
				}

				dBeta += Math.acos( dCosAlpha ); //* 1/sqrt(2)????
				oPosition.rotateY( dBeta*180.0/Math.PI );
				
				if( bModeSquare )
				{
					dLengthSquare = CalcSpiralLengthSquare( oPosition.getZ(), dLastRadius, dActualRadius );
					dCosAlpha = CalcSpiralAlpha( dLastRadius, dActualRadius, dLengthSquare );
					//dCosBeta = CalcSpiralBeta( dLastRadius, dActualRadius, dLengthSquare );			
				}
				else
				{
					dLengthSquare = CalcSpiralLength( dLastRadius, dActualRadius );
					dCosAlpha = CalcSpiralAlpha( dLastRadius, dActualRadius, dLengthSquare );
					//dCosBeta = CalcSpiralBeta( dLastRadius, dActualRadius, dLengthSquare );			
				}

				dGamma += Math.acos( dCosAlpha ); //* 1/sqrt(2)????
				oPosition.rotateZ( dGamma*180.0/Math.PI );

				//Vector3d oDirection = oPosition.div( oPosition.AbsoluteValue() ); //set direction unity-vector
			}

			oPosition.setZ( oPosition.getZ() + dShiftZ ); //parallel spiral with distance to center = dShiftZ

			//Draw Spiral
			//Mirror spiral if nIndex is set.
			if( Math.abs(oPosition.getX())<0.02 || dShiftZ==0.0 )
			{
				if( _sim.paramCheckColor ) //use rainbow-colors?
				{
					double dAbs = ( oPosition.sub(oLastPosition) ).absoluteValue();
					gl.glColor4d( 0.5 + (oPosition.getX()-oLastPosition.getX()) / dAbs
								, 0.5 + (oPosition.getY()-oLastPosition.getY()) / dAbs
								, 0.5 + (oPosition.getZ()-oLastPosition.getZ()) / dAbs, 1.0); //X
				}

				oPosition.mulEqu( 1.0 ); //cnGridSize;
				if( nIndex==0 ) gl.glVertex3d(oPosition.getX(), oPosition.getY(), oPosition.getZ() );
				if( nIndex==1 ) gl.glVertex3d(-oPosition.getX(), -oPosition.getY(), -oPosition.getZ() );

				if( nIndex==2 ) gl.glVertex3d(-oPosition.getZ(), oPosition.getX(), -oPosition.getY() );
				if( nIndex==3 ) gl.glVertex3d(oPosition.getZ(), -oPosition.getX(), oPosition.getY() );

				if( nIndex==4 ) gl.glVertex3d(-oPosition.getY(), -oPosition.getZ(), -oPosition.getX() );
				if( nIndex==5 ) gl.glVertex3d(oPosition.getY(), oPosition.getZ(), oPosition.getX() );
			}

			oLastPosition = oPosition; //save the last position
			dLastRadius = dActualRadius; //save the last radius-state.
		}
		gl.glEnd();
		
		if( _sim.paramFunction )
		{
			gl.glBegin( GL.GL_LINE_STRIP ); //Lemniskate - Lissajous figur
			gl.glColor4d(1.0, 1.0, 1.0, 1.0); //white
			for( double t=0; t<10.0; t+=0.1 )
			{
				double x = (t*t-1) / (t*t+1);
				double y = 2*t*(t*t-1) / (t*t+1) / (t*t+1);
				gl.glVertex3d( 0.0, y, x );
			}
			gl.glEnd();

			gl.glBegin( GL.GL_LINE_STRIP ); //Parabola
			for( double x=-1.0; x<1.0; x+=0.1 )
				gl.glVertex3d( 1-x*x, 0.0, x );
			gl.glEnd();
		}
	}

	/**
	 * 6 spiral-arms in every orthogonal direction
	 */
	void OnSpiral3D_SixSides( boolean bModeSquare, boolean constRadius )
	{
		//double dShiftZ=0.0;
		for( int nIndex=0; nIndex<6; nIndex++ )
		{
			OnSpiral3D( bModeSquare, false, nIndex, constRadius );
		}
	}
	
	/**
	 * draw spirals
	 */
	public void display( GLAutoDrawable gLDrawable )
	{
		gl = gLDrawable.getGL();
		//glu = new GLU();
		gl.glClearColor( 0.3f, 0.3f, 0.3f, 0f );
		gl.glClear( GL.GL_COLOR_BUFFER_BIT | GL.GL_DEPTH_BUFFER_BIT );
		gl.glDisable( GL.GL_LIGHTING ); //lines do not have a lightning-direction
	    gl.glDisable( GL.GL_DEPTH_TEST ); //sort the faces from back to front
		gl.glDisable( GL.GL_CULL_FACE ); //select only winding-correct

		GLPespective(); //TODO? call here? Needed for zoom-Keys

		gl.glPushMatrix(); //push parameter for the next loop
		gl.glTranslated(_PositionX, _PositionY, _PositionZ);
		gl.glRotated( _tetha, 1.0,  0.0,  0.0 );
		gl.glRotated( _phi, 0.0,  1.0,  0.0 );

		drawCube( 1.0 );
		
		if( _sim.bSpiral2D ) OnSpiral2D( false );
		if( _sim.bSpiral3D ) OnSpiral3D( false, _sim.bMode2D, 0, _sim.paramConstRadius );
		if( _sim.bSpiral6S ) OnSpiral3D_SixSides( false, _sim.paramConstRadius );

		if( _sim.bSpiral2DS ) OnSpiral2D( true );
		if( _sim.bSpiral3DS ) OnSpiral3D( true, _sim.bMode2D, 0, _sim.paramConstRadius );
		if( _sim.bSpiral6SS ) OnSpiral3D_SixSides( true, _sim.paramConstRadius );

		//TODO DrawCornerCoodinateSystem();


		//G4-2er-Orbital of hydrogen has the same form than a rotation around 2-axes (45 degree-view). 		
//		gl.glBegin( GL.GL_LINE_STRIP );
//		gl.glColor4d(1.0, 1.0, 1.0, 1.0); //white
//		
//		double dBeta=0.0; //spiral rotation around z/y-axis
//		for( int i=0; i<100; i++ ) 
//		{
//			Vector3d oPosition = new Vector3d( 1.0, 0.0, 0.0 );
//		
//			dBeta +=  0.1;
//			oPosition.RotateY( dBeta*180.0/Math.PI );
//			oPosition.RotateZ( dBeta*180.0/Math.PI );
//		
//			gl.glColor4d( 0.5 + (oPosition.getX()) , 0.5 + (oPosition.getY()) , 0.5 + (oPosition.getZ()), 1.0); //X
//			gl.glVertex3d(oPosition.getX(), oPosition.getY(), oPosition.getZ() );
//		}
//		gl.glEnd();
		
		
		gl.glPopMatrix(); //rotate back for rotation independent working tools like ruler or selectable points
		gl.glFlush();
	}

	@Override
	public void drawPickingObjects(GL gl)  { } //not used yet.	
//	@Override public void dispose(GLAutoDrawable arg0) { }
}



